/* $Workfile: T2mRawData.c $       $Revision: 31 $      $Modtime: 12/15/00 2:27p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2MRAWDATA.C

    A simple example console application demonstrating the use of the 
    Raw Data Sender and the Raw Data Receiver with the OTX Driver using the 
    Thor-2-PCMCIA board.

    The Demo application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 2 E1 Lines and expects them to be looped with an external
    loop cable. The LI Internal HDLC controller is used to send HDLC frames from
    one Line Interface to be received by the other.

    The Raw Data Receiver and the Raw Data Sender is implemented in the
    DSP device.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxRaw.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxBrd.h"

#define DEMO_DEFAULT_LOG_FILE "T2mRawData.log"
#define DEMO_DEFAULT_BOARD_NO 0

//#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call

#define DEMO_RAW_TX_BUF_SIZE 1024
//#define DEMO_RAW_TX_BUF_SIZE 2000

#define DEMO_TEST_STRING "abcdefg"

#define DEMO_RAW_RECEIVER 1

#define DEMO_DSP_MAX 2 // For debugging of OtxHdlc using CodeComposer set DEMO_DSP_MAX to 4

//#define DEMO_LINK_WITH_LIBCMT 1

#ifdef DEMO_LINK_WITH_LIBCMT
#include <process.h>
#endif

typedef struct DemoEnvT {
  char           *logFileName;
  OtxPhysicalDeviceTypeE nBoardType;
  OtxDataFormatE  law;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
  int             boardNo;
  OTX_INT32       nRawDataCapacity;
  OtxDspSpmE      eProgramModule;      // Standard Program Module to be loaded in the DSP (from the Lib)
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoCreatePhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
OTX_INT32   g_cbLiHdlcSenderFrame  = 10;
OTX_INT32   g_cbDspHdlcSenderFrame = 5;
OTX_INT32   g_coHdlcSenderFrames   = 1;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_UINT32  g_coHdlcRxFrames[32];
OTX_UINT32  g_coHdlcBadRxFrames[32];


//CRITICAL_SECTION g_analyzeCritSec;
//CRITICAL_SECTION g_demoLogfCritSec;


struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hVidarAsmController0;
  OTX_HANDLE hLi[2];
  OTX_HANDLE hDsp[DEMO_DSP_MAX];
} g_hPhysDevs;



struct {
  // External Pipes (The T1/E1 span) 
  OTX_HANDLE hLiIn;
  OTX_HANDLE hLiOut;

  // Pipes for HDLC Sender/Receiver in the DSP
  OTX_HANDLE hDspLiHdlc;
  OTX_HANDLE hLiDspHdlc;

  // Pipes for Raw Data Sender/Receiver in the DSP
  OTX_HANDLE hDspLiRawData;
  OTX_HANDLE hLiDspRawData;
} g_hPipes[2];


struct sRawDataSender {
  OTX_HANDLE hHandle;
  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_UINT32 nSourceId;

  //DWORD nThreadId;
  //HANDLE hThread;

  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_RAW_TX_BUF_SIZE];
  FILE       *pTxFile;
  OTX_INT32   cbTxBuf;
  OTX_INT32   coUnderRunErrors;
};

struct sRawDataReceiver {
  OTX_HANDLE hHandle;
  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;
  OTX_UINT32 nSourceId;

  //DWORD nThreadId;
  //HANDLE hThread;

  OTX_BOOL    bVerbose;
  OTX_BOOL    bActive;
  FILE       *pRxFile;
};

struct {
  OTX_HANDLE hHdlcSender[2];
  OTX_HANDLE hHdlcReceiver[2];
  OTX_HANDLE hDspHdlcSender[2];
  OTX_HANDLE hDspHdlcReceiver[2];
  struct sRawDataSender mRawDataSender[2];
#ifdef DEMO_RAW_RECEIVER
  struct sRawDataReceiver mRawDataReceiver[2];
#endif
} g_hLogicDevs;

struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;




/*******************************************************************************
*******************************************************************************/


void DemoCreateEventQueue(void)
/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hEventQueue, result);

}



void DemoCreatePhysicalDevices(void)
/*******************************************************************************

  @func

  Create and Connect the Physical Devices on the Thor-2 board

  @comm

    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

    The physical devices can be classified in 2 categories:

  - BOARD Devices: A board device represents one OTX Network Interface card or
    a Daughter Board Module. The board device is the place holder and controller
    of other physical and sometime logical devices representing
    the devices on the board: such as Line Interface and Filter  Devices.


  - CHIP Devices: A chip device represents a physical integrated circuit
    on the board performing one or several functions. A logical device can
    be created to represent each of the functions implemented by a chip device.
    For example, a Digital Signal Processor is a physical "chip" device.
    However, the DSP software may implement several functions and be
    represented by a multiple of logical devices.



*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_THOR_PCI:
      DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
      break;
    case OTX_DEVICE_THOR_PCMCIA:
      DemoLogf("Opening OTX_DEVICE_THOR_PCMCIA board\n");
      break;
    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      OsaConsoleExit();
      exit(1);
      break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);


  // Open the two Line interfaces
  for (i = 0; i < 2; i++) {
    // Get handles to T1/E1 line interface devices #0 and #1
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }


  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCMCIA) {
    // Nothing
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_PCI) {

    // Get a handle to to the ASM Daughter Board
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_VIDAR_5x_ASM, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    
  } else {
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    OsaConsoleExit();
    exit(1);
  }
  
  // Get handles to the physical DSP devices
  for (i = 0; i < DEMO_DSP_MAX; i++) {
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }
  
  // Load DSP Program
  for (i = 0; i < DEMO_DSP_MAX; i++) {
    
    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXSPM1:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXSPM1\n", i);
      break;
    case OTX_DSP_SPM_OTXHDLC:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXHDLC\n", i);
      break;
    default:
      DemoLogf("Unsupport DSP program module (%d) for this demo program. Exiting.\n", g_demoEnv.eProgramModule);
      OtxDrvReleaseLib();
      exit(1);
    }
    
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], g_demoEnv.eProgramModule); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  for (i = 0; i < 2; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    // Configure Transceivers
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);

  }

}



void DemoEnablePhysicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  // Enable the board (Physical Device)
  DemoLogf("Enable hThorController0\n");
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
  }

  for (i = 0; i < 2; i++) {

    // Enable Physical Devices
    DemoLogf("Enable g_hPhysDevs.hLi[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hLi[i],  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    DemoLogf("g_hPhysDevs.hDsp[%d]\n", i);
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);

  }

}


void DemoCreateLogicalDevices(
/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

    Logical Devices can be classified into Control Device and Data Devices:

    - Control Device:  A logical device performing a control function. E.g. Event Queue.

    - Data Filter Device: A device implemented in Hardware, Firmware of Software
    which converts, processes, or manipulates serial data. Examples of data devices include
    Time-Space Switch, HDLC Controller, Tone Detector, Tone Generator, etc.

    Data Filter devices can be categorized into 3 basic types:

     - Data Source: Originates a Pipe or adds data to an existing Pipe. E.g. DTMF Tone Generator.

     - Data Sink: Terminates a Pipe or extracts data from a Pipe. E.g. DTMF Detector.

     - Data Bridge: Cross-connects data from one Pipe to another or combines data from
     multiple Pipes into one Pipe. E.g. cross-connect in TSS.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;

  // LI#0 and LI#1: Create Hdlc Senders and Receivers logical devices hosted in the physical Line Interface devices
  for (i = 0; i < 2; i++) {
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hLi[i], OTX_LDEVICE_HDLC_SENDER,   g_hEventQueue, i, &g_hLogicDevs.hHdlcSender[i]); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hLi[i], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_hLogicDevs.hHdlcReceiver[i]); DemoCheckRc(g_hPhysDevs.hLi[i], result);
    //g_sqDspHdlcReceivers[g_hLogicDevs.hHdlcReceiver0 & 0x1F].hHandle = g_hLogicDevs.hHdlcReceiver0;
  }

  // DSP#0 and DSP#1: Create Hdlc Senders and Receivers logical devices hosted in the physical DSP devices
  for (i = 0; i < 2; i++) {
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_HDLC_SENDER,   g_hEventQueue, 2+i, &g_hLogicDevs.hDspHdlcSender[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, 2+i, &g_hLogicDevs.hDspHdlcReceiver[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    //g_sqDspHdlcReceivers[g_hLogicDevs.hHdlcReceiver0 & 0x1F].hHandle = g_hLogicDevs.hHdlcReceiver0;
  }

  for (i = 0; i < 2; i++) {
    // Create a notification Event and Event Queue for the Raw Data Sender
    result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataSender[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_hLogicDevs.mRawDataSender[i].hNotificationEvent, &g_hLogicDevs.mRawDataSender[i].hEventQueue); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hEventQueue, result);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hEventQueue, result);

    // Create Raw Data Sender
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_DATA_RAW_SENDER, g_hLogicDevs.mRawDataSender[i].hEventQueue, i, &g_hLogicDevs.mRawDataSender[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    g_hLogicDevs.mRawDataSender[i].nSourceId = i; // Store sourceId so that we can easily print it from the Sender Thread

    // Create a notification Event and Event Queue for the Raw Data Receiver
#ifdef DEMO_RAW_RECEIVER
    result = OtxDrvCreateEvent(&g_hLogicDevs.mRawDataReceiver[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvCreateEventQueue(128, g_hLogicDevs.mRawDataReceiver[i].hNotificationEvent, &g_hLogicDevs.mRawDataReceiver[i].hEventQueue); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hEventQueue, result);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hEventQueue, result);

    // Create Raw Data Receiver
    result = OtxDrvCreateLogicalDevice(g_hPhysDevs.hDsp[i], OTX_LDEVICE_DATA_RAW_RECEIVER, g_hLogicDevs.mRawDataReceiver[i].hEventQueue, i, &g_hLogicDevs.mRawDataReceiver[i].hHandle); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    g_hLogicDevs.mRawDataReceiver[i].nSourceId = i; // Store sourceId so that we can easily print it from the Receiver Thread
#endif
  }


}


void DemoConnectDevices(
/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];



  // Create pipes for the HDLC Logicial Device in the Line Interface (one for each direction)
  for (i = 0; i < 2; i++) {
    OTX_INT32 nCapacity;
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    
#if 0
    //memset(sqTimeSlotMask+24, 0xFF, 8 * sizeof(OTX_UINT8));
    //nCapacity = 8*64;
    sqTimeSlotMask[16] = 0xFF; // Timeslot 16
    sqTimeSlotMask[17] = 0xFF; // Timeslot 17
    nCapacity = 2*64;
#else
	{
    OTX_INT32 ts;
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      ts = 24; 
    } else {
      // E1 mode
      ts = 16; 
    }
    
    sqTimeSlotMask[ts] = 0xFF;
    nCapacity = 64;
	}
#endif      
    
    // Create an incoming and outgoing Pipe representing the external T1/E1
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[i].hLiIn); DemoCheckRc(g_hPipes[i].hLiIn, result);
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[i].hLiOut); DemoCheckRc(g_hPipes[i].hLiOut, result);
    
    // Connect external T1/E1 pipes. Only connect one end of each pipe. External Interface is LI Highway Number 1
    result = OtxDrvConnectPipe(g_hPipes[i].hLiIn,  OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hLiIn, result);
    result = OtxDrvConnectPipe(g_hPipes[i].hLiOut, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[i], 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hLiOut, result);
  }


  // CONNECT LOGICAL DEVICES WITHIN THE LIs
  // --------------------------------------
  for (i = 0; i < 2; i++) {
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hHdlcSender[i], -1, g_hPipes[i].hLiOut, -1); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);
    result = OtxDrvConnectLogicalDevice(g_hPipes[i].hLiIn,           -1, g_hLogicDevs.hHdlcReceiver[i], -1); DemoCheckRc(g_hPipes[i].hLiIn, result);
  }




  // Create the Pipes (one for each direction) between the two DSPs and the two LIs
  for (i = 0; i < 2; i++) {
    OTX_INT32 nCapacity;
    OTX_INT32 nTs;
    
    // Create HDLC pipes (one for each direction) for timeslot 1 (and timeslot 2 if Multi-slot pipe) 
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    
#if 0
    sqTimeSlotMask[1] = 0xFF; // Timeslot 1
    sqTimeSlotMask[2] = 0xFF; // Timeslot 2
    nCapacity = 2*64;
#else
    sqTimeSlotMask[1] = 0xFF; // Timeslot 1
    nCapacity = 64;
#endif      
    
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[i].hLiDspHdlc); DemoCheckRc(g_hPipes[i].hLiDspHdlc, result);
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[i].hDspLiHdlc); DemoCheckRc(g_hPipes[i].hDspLiHdlc, result);
    
    // Connect pipe: Dsp -> Li 
    result = OtxDrvConnectPipe(g_hPipes[i].hDspLiHdlc, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hDspLiHdlc, result);
    result = OtxDrvConnectPipe(g_hPipes[i].hDspLiHdlc, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hDspLiHdlc, result);
    // Connect pipe: Li  -> Dsp 
    result = OtxDrvConnectPipe(g_hPipes[i].hLiDspHdlc, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hLiDspHdlc, result);
    result = OtxDrvConnectPipe(g_hPipes[i].hLiDspHdlc, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hLiDspHdlc, result);
    


    // Create Raw Data pipes (one for each direction) for timeslot 9 (and timeslot 10 if Multi-slot pipe) 
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    
    if (g_demoEnv.nRawDataCapacity >= 64) {
      // Standard (64kpbs) or superchannel
      nCapacity = 0;
      for (nTs = 0; nTs < 32; nTs++) {
        
        if (nCapacity == g_demoEnv.nRawDataCapacity) {
          break;
        }
        
        if (nTs == 0) {
          continue; // TS0 is sync timeslot
        }
        if (nTs == 1) {
          continue; // TS1 used by the DSP HDLC Sender/Receiver
        }
        if (nTs == 16) {
          continue; // TS16 is used by the T1E1 HDLC Sender/Receiver 
        }
        
        sqTimeSlotMask[nTs] = 0xFF; 
        nCapacity += 64;
      }
      
      if (nCapacity != g_demoEnv.nRawDataCapacity) {
        DemoLogf("Error: Could not connect Raw Data Sender/Receiver with capacity=%d\n", g_demoEnv.nRawDataCapacity);
        OsaConsoleExit();
        OtxDrvReleaseLib();
        exit(1);
      }


    } else {
      // Subchannel
      nTs = 2;
      switch (g_demoEnv.nRawDataCapacity) {
      case 8:
        sqTimeSlotMask[nTs] = 0x01;
        break;
      case 16:
        sqTimeSlotMask[nTs] = 0x03;
        break;
      case 32:
        sqTimeSlotMask[nTs] = 0x0F;
        break;
      default:
        DemoLogf("Error: Unsupported capacity=%d\n", g_demoEnv.nRawDataCapacity);
        OsaConsoleExit();
        OtxDrvReleaseLib();
        exit(1);
        break;
        
      }
      nCapacity = g_demoEnv.nRawDataCapacity;
    }
    
    
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[i].hLiDspRawData); DemoCheckRc(g_hPipes[i].hLiDspRawData, result);
    result = OtxDrvCreatePipe(nCapacity, &g_hPipes[i].hDspLiRawData); DemoCheckRc(g_hPipes[i].hDspLiRawData, result);
    
    // Connect pipe: Dsp -> Li 
    result = OtxDrvConnectPipe(g_hPipes[i].hDspLiRawData, OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hDspLiRawData, result);
    result = OtxDrvConnectPipe(g_hPipes[i].hDspLiRawData, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hDspLiRawData, result);
    // Connect pipe: Li  -> Dsp 
    result = OtxDrvConnectPipe(g_hPipes[i].hLiDspRawData, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hLiDspRawData, result);
    result = OtxDrvConnectPipe(g_hPipes[i].hLiDspRawData, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[i], -1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes[i].hLiDspRawData, result);
    
  }



  // CONNECT LOGICAL DEVICES WITHIN THE DSP
  // --------------------------------------
 for (i = 0; i < 2; i++) {
    // Receive direction:
    // Connect HDLC Receiver
    result = OtxDrvConnectLogicalDevice(g_hPipes[i].hLiDspHdlc,     -1,  g_hLogicDevs.hDspHdlcReceiver[i],         -1); DemoCheckRc(g_hPipes[i].hLiDspHdlc, result);
    // Connect Raw Data Receiver
#ifdef DEMO_RAW_RECEIVER
    result = OtxDrvConnectLogicalDevice(g_hPipes[i].hLiDspRawData,  -1,  g_hLogicDevs.mRawDataReceiver[i].hHandle, -1); DemoCheckRc(g_hPipes[i].hLiDspRawData, result);
#endif
    // Transmit direction:
    // Connect HDLC Sender
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.hDspHdlcSender[i],         -1, g_hPipes[i].hDspLiHdlc,    -1); DemoCheckRc(g_hLogicDevs.hDspHdlcSender[i], result);
    // Connect Raw Data Sender
    result = OtxDrvConnectLogicalDevice(g_hLogicDevs.mRawDataSender[i].hHandle, -1, g_hPipes[i].hDspLiRawData, -1); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);

  }

 
}


void DemoInitializeLogicalDevices(
/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after 
    the devices are created and connected.

*******************************************************************************/
  void
)
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < 2; i++) {

    // Initialize the Hdlc Sender and Receiver in the Lis
    result = OtxHdlcSenderInit(
      g_hLogicDevs.hHdlcSender[i], // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
      2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
      OTX_HDLC_TR_NA             // HDLC Transmission Rate for V.110 and X.30 modes
      ); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);          
    
    result = OtxHdlcReceiverInit(
      g_hLogicDevs.hHdlcReceiver[i], // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
      2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
      OTX_HDLC_TR_NA,            // HDLC Transmission Rate for V.110 and X.30 modes
      OTX_TRUE                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
      ); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);          
    
    // Initialize the Hdlc Sender and Receiver in the Dsps
    result = OtxHdlcSenderInit(
      g_hLogicDevs.hDspHdlcSender[i], // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
      2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
      OTX_HDLC_TR_NA             // HDLC Transmission Rate for V.110 and X.30 modes
      ); DemoCheckRc(g_hLogicDevs.hDspHdlcSender[i], result);          
    
    result = OtxHdlcReceiverInit(
      g_hLogicDevs.hDspHdlcReceiver[i], // Handle to logical HDLC Filter Device
      OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
      2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
      OTX_HDLC_TR_NA,            // HDLC Transmission Rate for V.110 and X.30 modes
      OTX_TRUE                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
      ); DemoCheckRc(g_hLogicDevs.hDspHdlcReceiver[i], result);          
    
    // Initialize the Raw Data Receivers (hosted in the DSP)
#ifdef DEMO_RAW_RECEIVER
    result = OtxRawDataReceiverInit(
      g_hLogicDevs.mRawDataReceiver[i].hHandle, 
      OTX_DATA_FORMAT_UINT8,  // No data conversion. Deliver the data as it is received on the highway
      0xFF,        // Flag to be filtered
      OTX_FALSE,  
      OTX_TRUE     // Least Significant part is received first. Only applicable for a subchannel pipe (nCapacity < 64).
      ); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
#endif
    // Initialize the Raw Data Senders (hosted in the DSP)
    result = OtxRawDataSenderInit(
      g_hLogicDevs.mRawDataSender[i].hHandle, 
      OTX_DATA_FORMAT_UINT8,  // No data conversion. 
      0xFF,        // Flag to be sent when the sender is out of data
      OTX_TRUE     // Least Significant part is transmitted first. Only applicable for a subchannel pipe (nCapacity < 64).
      ); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);
  }

}


void DemoEnableLogicalDevices(void)
/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
{
  OTX_RESULT result;
  OTX_INT32 i;

  for (i = 0; i < 2; i++) {

    DemoLogf("Enable hHdlcSender[%d]\n", i);
    result = OtxDrvEnable(g_hLogicDevs.hHdlcSender[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcSender[i], result);

    DemoLogf("Enable hHdlcReceiver[%d]\n", i);
    result = OtxDrvEnable(g_hLogicDevs.hHdlcReceiver[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcReceiver[i], result);

    DemoLogf("Enable hDspHdlcSender[%d]\n", i);
    result = OtxDrvEnable(g_hLogicDevs.hDspHdlcSender[i],   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDspHdlcSender[i], result);

    DemoLogf("Enable hDspHdlcReceiver[%d]\n", i);
    result = OtxDrvEnable(g_hLogicDevs.hDspHdlcReceiver[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hDspHdlcReceiver[i], result);

    DemoLogf("Enable mRawDataSender[%d]\n", i);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataSender[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataSender[i].hHandle, result);

    
#ifdef DEMO_RAW_RECEIVER
#if 0 // We enable the Receiver in the DemoRawDataReceiver thread instead.
    OsaThreadSleep(100);
    DemoLogf("Enable mRawDataReceiver[%d]\n", i);
    result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[i].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[i].hHandle, result);
#endif
#endif
  }

}


void DemoInitT1E1ConfigData(
/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_FALSE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 128;
  pE1CfgData->nLosRecovery               = 128;
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 128;
  pT1CfgData->nLosRecovery               = 128;
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x00;
}



static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  short c;
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d): %s \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      } else {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d) Cannot translate this result code \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber);
      }
    }


    DemoLogf("Do you want to exit (Y/N) ");
    c = getch();
    if (c == 'y' || c == 'Y') {
      // Clean up
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }

  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct tm *ptCurrentDayAndTime;
  time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with localtime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = localtime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
    
}


/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (result == OTX_S_OK);
  }

  return 0;

}

#ifdef DEMO_RAW_RECEIVER
/*******************************************************************************
  @func
    Raw Data Receiver Thread Function 
*******************************************************************************/
#ifdef DEMO_LINK_WITH_LIBCMT
void DemoRawDataReceiver(
  void *lpParameter   // thread data
)
#else
void *DemoRawDataReceiver(
  void *lpParameter   // thread data
)
#endif
{
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  OTX_UINT8  buf[2048];
  OTX_UINT32 nFrameLen;
  OTX_BOOL bFrameEnd;
  char msgStr[256];
  clock_t startTime, stopTime;
  OTX_UINT32 nSourceId;

  startTime = clock();

  nSourceId = *((OTX_INT32 *) lpParameter);
  if (nSourceId >= 2) {
    DemoLogf("DemoRawDataReceiver: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    goto FATALEND;
  }

  g_hLogicDevs.mRawDataReceiver[nSourceId].bVerbose = OTX_FALSE;
  g_hLogicDevs.mRawDataReceiver[nSourceId].bActive  = OTX_TRUE;

  // Enable the receiver
  result = OtxDrvEnable(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, result);

  // Discard any events from event queue (if any)
  do {
    result = OtxDrvGetEventData(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, result);
  } while (result == OTX_S_OK);
  // Discard any events from event queue (if any)
  do {
    result = OtxRawDataReadData(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, buf, 2048, &nFrameLen, &bFrameEnd);
  } while (result == OTX_S_OK);
  if (result != OTX_S_FALSE) {
    DemoLogf("RawDataReceiver[%d] OtxRawDataReadData() was unsuccessful.\n", nSourceId);
    DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, result);
  }



  while (g_hLogicDevs.mRawDataReceiver[nSourceId].bActive) {
    OtxDrvWaitForSingleEvent(g_hLogicDevs.mRawDataReceiver[nSourceId].hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("RawDataReceiver[%d] Received Event\n");
    do {
      result = OtxDrvGetEventData(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        switch (eventData.m_eDeviceType) {
        case OTX_LDEVICE_DATA_RAW_RECEIVER:
          // Event from Raw Data Receiver 
          if ((eventData.m_nSourceId != nSourceId)) {
            DemoLogf("RawDataReceiver[%d] Error: m_nSourceId is out of range.\n", eventData.m_nSourceId);
            break;
          } else {
             i = eventData.m_nSourceId;
          }

          switch (eventData.m_nCode) {
          case OTX_RAWDATA_RECEIVER_EC_DATA_AVAILABLE:
            {
              otxRc = OtxRawDataReadData(eventData.m_hDevice, buf, 2048, &nFrameLen, &bFrameEnd);
              while (otxRc == OTX_S_OK) {
                OTX_INT32 x;
                
                if (g_hLogicDevs.mRawDataReceiver[i].bVerbose) {
                  DemoLogf("RawDataReceiver[%d] (len=%.4d) ", eventData.m_nSourceId, nFrameLen);
                  for (x = 0; x < 16; x++) {
                    DemoLogf("%.2X ", buf[x]);
                  }
                  DemoLogf("\n");
                }

                if (g_hLogicDevs.mRawDataReceiver[i].pRxFile) {
                  int cbWritten;

                  cbWritten = fwrite(buf, sizeof(OTX_UINT8), nFrameLen, g_hLogicDevs.mRawDataReceiver[i].pRxFile);

                  if (cbWritten != (int)nFrameLen) {
                    DemoLogf("RawDataReceiver[%d] Error: Could not write data to file.  (nFrameLen=%d, cbWritten=%d) \n", eventData.m_nSourceId, nFrameLen, cbWritten);
                  }
                  
                  if (bFrameEnd) {
                    DemoLogf("RawDataReceiver[%d] Data reception complete. Closing data file. \n", eventData.m_nSourceId);
                    fclose(g_hLogicDevs.mRawDataReceiver[i].pRxFile);
                    g_hLogicDevs.mRawDataReceiver[i].pRxFile = NULL;
                    goto END;
                  }
                }
                
                otxRc = OtxRawDataReadData(eventData.m_hDevice, buf, 2048, &nFrameLen, &bFrameEnd);
              }
              if (otxRc != OTX_S_FALSE) {
                DemoLogf("RawDataReceiver[%d] OtxRawDataReadData() was unsuccessful.\n", eventData.m_nSourceId);
                DemoCheckRc(eventData.m_hDevice, otxRc);
              }

            }
            break;
          default:
            otxRc = OtxDrvEventCode2String(eventData.m_hDevice, eventData.m_nCode, 256, &msgStr[0]);
            // Unexpected event from this device
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("RawDataReceiver[%d] Event: %s \n", eventData.m_nSourceId, msgStr);
            } else {
              DemoLogf("RawDataReceiver[%d] Invalid Event: %ld \n", eventData.m_nSourceId, eventData.m_nCode);
            }
            break;
          }
          break;
          default:
            // Unhandled event
            DemoLogf("DemoRawDataReceiver[%d]: Event from an unknown object 0x%X\n", nSourceId, eventData.m_hDevice);
            break;
        }
        
      }
    } while (result == OTX_S_OK);

    OsaThreadSleep((256*64)/g_demoEnv.nRawDataCapacity); // Give the driver some time to fill up more data for us to fetch (256ms fills 2048 bytes in a 64kbit/s pipe)

  }

END:
  
  stopTime = clock();

  // Disable the receiver
  result = OtxDrvDisable(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.mRawDataReceiver[nSourceId].hHandle, result);
  
  DemoLogf("DemoRawDataReceiver[%d]: Done (time=%2.1f seconds).\n", nSourceId, (double)(stopTime - startTime) / CLOCKS_PER_SEC);

FATALEND:
#ifdef DEMO_LINK_WITH_LIBCMT
  _endthread();
#else
  return 0;
#endif

}
#endif

/*******************************************************************************
  @func
    Raw Data Sender Thread Function 
*******************************************************************************/

#ifdef DEMO_LINK_WITH_LIBCMT
void DemoRawDataSender(
  void *lpParameter   // thread data
)
#else
void *DemoRawDataSender(
  void *lpParameter   // thread data
)
#endif
{
  OtxEventDataS eventData;
  OTX_RESULT result;
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  char msgStr[256];
  clock_t startTime, stopTime;
  OTX_UINT32 nSourceId;
  OTX_UINT32 nTaskRef = 0;
  

  startTime = clock();

  nSourceId = *((OTX_INT32 *) lpParameter);
  if (nSourceId >= 2) {
    DemoLogf("DemoRawDataSender: Thread was passed an invalid nSourceId(%d). Thread exiting...\n", nSourceId);
    goto END;
  }

  g_hLogicDevs.mRawDataSender[nSourceId].bActive = OTX_TRUE;
  g_hLogicDevs.mRawDataSender[nSourceId].coUnderRunErrors = 0;

  // Discard any events from event queue (if any)
  do {
    result = OtxDrvGetEventData(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, result);
  } while (result == OTX_S_OK);

  if (g_hLogicDevs.mRawDataSender[nSourceId].pTxFile) {
    // Send from a file
    do {
      g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf = fread(g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, sizeof(OTX_UINT8), DEMO_RAW_TX_BUF_SIZE, g_hLogicDevs.mRawDataSender[nSourceId].pTxFile);
      
      otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
      if (otxRc != OTX_E_BUSY) {
        DemoLogf("DemoRawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
        nTaskRef++;
      }
      
    } while (g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf && (otxRc != OTX_E_BUSY));
    
    if (g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf) {
      //DemoLogf("DemoRawDataSender[%d]: Waiting for OTX_RAWDATA_SENDER_EC_BUF_DEPLETED events to complete the transmission...\n", nSourceId);
    } else {
      DemoLogf("DemoRawDataSender[%d]: Transmission complete. Closing file\n", nSourceId);
      fclose(g_hLogicDevs.mRawDataSender[nSourceId].pTxFile);
      g_hLogicDevs.mRawDataSender[nSourceId].pTxFile = NULL;
      goto END;
    }
  } else {
    // Send a text string (continously until aborted)
    OTX_INT32 cbText = strlen(DEMO_TEST_STRING);
    OTX_INT32 j;

    DemoLogf("DemoRawDataSender[%d]: Preparing text string buffer...\n", nSourceId);
    
    g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf[0] = '\0';

    // Fill sqTxBuf with the text string (as many copies of the text string that will fit in sqTxBuf). 
    for (j = 0; j < (DEMO_RAW_TX_BUF_SIZE-1)/cbText; j++) {
      strcpy((char *)g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf + strlen((char *)g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf), DEMO_TEST_STRING);
    }

    g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf = strlen((char *)g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf);

    DemoLogf("DemoRawDataSender[%d]: Done preparing text string buffer (len=%d)...\n", nSourceId, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf);

    // Prime the sender
    do {
      otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[nSourceId].hHandle, g_hLogicDevs.mRawDataSender[nSourceId].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
      if (otxRc != OTX_E_BUSY) {
        //DemoLogf("DemoRawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
        nTaskRef++;
      }

    } while (otxRc != OTX_E_BUSY);
     
  }
  
  
    while (g_hLogicDevs.mRawDataSender[nSourceId].bActive) {
    OtxDrvWaitForSingleEvent(g_hLogicDevs.mRawDataSender[nSourceId].hNotificationEvent, OTX_TIME_INFINITY);
    do {
      result = OtxDrvGetEventData(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, &eventData); DemoCheckRc(g_hLogicDevs.mRawDataSender[nSourceId].hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        switch (eventData.m_eDeviceType) {
        case OTX_LDEVICE_DATA_RAW_SENDER:
          // Event from Raw Data Sender 
          if ((eventData.m_nSourceId != nSourceId)) {
            DemoLogf("RawDataSender[%d] Error: m_nSourceId is out of range.\n", eventData.m_nSourceId);
            break;
          } else {
             i = eventData.m_nSourceId;
          }

          switch (eventData.m_nCode) {
          case OTX_RAWDATA_SENDER_EC_COMPLETED:
            if (!g_hLogicDevs.mRawDataSender[i].pTxFile) {
              g_hLogicDevs.mRawDataSender[i].coUnderRunErrors++;
              //DemoLogf("DemoRawDataSender[%d]: *** Data Underrun at cbTxSent/cbTxBuf=%d/%d (m_nParam=%d). \n", nCin, g_sqCin[nCin].cbTxSent, g_sqCin[nCin].cbTxBuf, eventData.m_nParam);
              // Fall-through
            } else {
              DemoLogf("DemoRawDataSender[%d]: Packet #%d sent. Transmission completed.\n", eventData.m_nSourceId, eventData.m_nRequestId);
              fclose(g_hLogicDevs.mRawDataSender[i].pTxFile);
              g_hLogicDevs.mRawDataSender[i].pTxFile = NULL;
              goto END;
            }
            // Fall-through
          case OTX_RAWDATA_SENDER_EC_BUF_DEPLETED:
            {
              OTX_RESULT otxRc;
              
              if (g_hLogicDevs.mRawDataSender[i].pTxFile) {
                // Sending from file
                DemoLogf("DemoRawDataSender[%d]: Packet #%d in progress\n", eventData.m_nSourceId, eventData.m_nRequestId);

                if (g_hLogicDevs.mRawDataSender[i].cbTxBuf) {
                  
                  while (g_hLogicDevs.mRawDataSender[i].cbTxBuf) {
                    otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[i].hHandle, g_hLogicDevs.mRawDataSender[i].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
                    if (otxRc == OTX_E_BUSY) {
                      break;
                    }
                    
                    DemoLogf("DemoRawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
                    nTaskRef++;
                    
                    g_hLogicDevs.mRawDataSender[i].cbTxBuf = fread(g_hLogicDevs.mRawDataSender[i].sqTxBuf, sizeof(OTX_UINT8), DEMO_RAW_TX_BUF_SIZE, g_hLogicDevs.mRawDataSender[nSourceId].pTxFile);
                  }
                  
                  if (!g_hLogicDevs.mRawDataSender[i].cbTxBuf) {
                    DemoLogf("DemoRawDataSender[%d]: EOF reached. Waiting for completion.\n", eventData.m_nSourceId);
                  }
                } else {
                  DemoLogf("DemoRawDataSender[%d]: Waiting for completion.\n", eventData.m_nSourceId);
                }
              } else if (g_hLogicDevs.mRawDataSender[i].cbTxBuf) {
                // Sending from static buffer (continously until aborted)
                while (g_hLogicDevs.mRawDataSender[nSourceId].bActive) {
                  otxRc = OtxRawDataSendData(g_hLogicDevs.mRawDataSender[i].hHandle, g_hLogicDevs.mRawDataSender[i].sqTxBuf, g_hLogicDevs.mRawDataSender[nSourceId].cbTxBuf, nTaskRef, OTX_TIME_INFINITY);
                  if (otxRc == OTX_E_BUSY) {
                    break;
                  }

                  //DemoLogf("DemoRawDataSender[%d]: Packet #%d buffered\n", nSourceId, nTaskRef);
                  nTaskRef++;
                }
              } else {
                // No file handles and no more bytes to send. Waiting for the OTX_RAWDATA_SENDER_EC_COMPLETED event
              }
            }
            break;
          
          
          default:
            otxRc = OtxDrvEventCode2String(eventData.m_hDevice, eventData.m_nCode, 256, &msgStr[0]);
            // Unexpected event from this device
            if (OTX_SUCCESS(otxRc)) {
              DemoLogf("RawDataSender[%d] Event: %s \n", eventData.m_nSourceId, msgStr);
            } else {
              DemoLogf("RawDataSender[%d] Invalid Event: %ld \n", eventData.m_nSourceId, eventData.m_nCode);
            }
            break;
          }
          break;
          default:
            // Unhandled event
            DemoLogf("DemoRawDataSender[%d]: Event from an unknown object 0x%X\n", nSourceId, eventData.m_hDevice);
            break;
        }
        
      }
    } while (result == OTX_S_OK);
  }

END:
  
  stopTime = clock();
  
  DemoLogf("DemoRawDataSender[%d]: Done (time=%2.1f seconds, coUnderRunErrors=%d).\n", nSourceId, (double)(stopTime - startTime) / CLOCKS_PER_SEC, g_hLogicDevs.mRawDataSender[nSourceId].coUnderRunErrors);
  
#ifdef DEMO_LINK_WITH_LIBCMT
  _endthread();
#else
  return 0;
#endif

}


void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  OTX_UINT8  buf[2048];
  OTX_UINT32 nFrameLen;
  OTX_UINT32 nFrameSeq;
  OtxHdlcFrameEndCodeE nFrameEndCode;
  OTX_DATETIME nTimeStamp;
  char msgStr[256];

  //EnterCriticalSection(&g_analyzeCritSec);


  DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    
  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
  case OTX_LDEVICE_HDLC_RECEIVER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from HLDC Receiver 
      if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
        
        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
        while (otxRc == OTX_S_OK) {

          if (pEventData->m_nSourceId < 32) {
            g_coHdlcRxFrames[pEventData->m_nSourceId]++;
          }

          if (g_bHdlcVerbose) {
            DemoLogf("HDLC Frame Received, Sequence Number = %d\n", pEventData->m_nParam);

            DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
            for (i = 0; i < nFrameLen && i < 1024; i++) {
              DemoLogf("%.2X ", buf[i]);
            }
            DemoLogf("\n");
          } else {
            if (nFrameEndCode) {

              g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
              
              DemoLogf("FrameEndCode Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
              
              if (g_coHdlcBadRxFrames[pEventData->m_nSourceId] > 100) {
                DemoLogf("Too many bad frames from HdlcReceiver[%d]. Disabling device\n", pEventData->m_nSourceId);
                otxRc = OtxDrvDisable(pEventData->m_hDevice,  OTX_TASK_SYNC, 5000L); DemoCheckRc(pEventData->m_hDevice, otxRc);
              }


            } else {
              if (nFrameLen != 10) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
                DemoLogf("LENGTH Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
                for (i = 0; i < nFrameLen && i < 256; i++) {
                  DemoLogf("%.2X ", buf[i]);
                }
                DemoLogf("\n");
                
              } else if (memcmp(buf, "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09", 10)) {
                g_coHdlcBadRxFrames[pEventData->m_nSourceId]++;
                
                DemoLogf("memcmp Error HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
                for (i = 0; i < nFrameLen && i < 256; i++) {
                  DemoLogf("%.2X ", buf[i]);
                }
                DemoLogf("\n");
                
              }
            }

          }
          otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
        }
        if (otxRc != OTX_S_FALSE) {
          DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
          DemoCheckRc(pEventData->m_hDevice, otxRc);
        }

      } else {
        // Unexpected event from this device
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


  //LeaveCriticalSection(&g_analyzeCritSec);

}

OTX_INT32 DemoProcessUserCommand(char commandChar)
/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
{
  static OTX_INT32 nCurrentLi = 0;
  OTX_RESULT result;

  DemoLogf("Key '%c' pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-1 Select active Li or Dsp, this Li/Dsp is then used for subsequent commands\n");
    DemoLogf(" x   Send a text file using the RawData Sender[%d]. Received by the RawData Receiver[%d]\n", nCurrentLi, !nCurrentLi);
    DemoLogf(" a   Abort the Raw Data Sender[%d] and Raw Data Receiver[%d]\n", nCurrentLi, !nCurrentLi);
    DemoLogf(" r   Start Raw Data Receiver[%d] (no file logging)\n", !nCurrentLi);
    DemoLogf(" s   Start Raw Data Sender[%d]. Prompt for file to send\n", nCurrentLi);
    DemoLogf(" S   Start Raw Data Sender[%d]. Continously send a text string\n", nCurrentLi);
    DemoLogf(" v   Toggle verbose (print received data) setting Raw Data Receiver[%d]\n", !nCurrentLi);
    DemoLogf(" l   Query Line Status of both Lis\n");
    DemoLogf(" t   Send a HDLC frame on TS16 (using HDLC Sender in the Li)\n");
    DemoLogf(" T   Send a HDLC frame on TS1  (using HDLC Sender in the DSP)\n");
    break;

#if 1 // Only for debugging
  case 'b':
    {
      char szBuf[12];
      OTX_INT32 nDsp;
      
      DemoLogf("Init Bsp. Which Dsp#?");
      OsaConsoleGetStr(szBuf);
      nDsp = (OTX_INT32) strtol(szBuf, NULL, 0);
      
      OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);

    }
    break;
  case 'B':
    {
      OTX_INT32 nDsp;
      
      for (nDsp = 0; nDsp < 2; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
    break;

  case 'f':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Forcing resynchronization on Li #%d... ", nLi);
        result = OtxT1E1ForceResynch(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;

  case 'F':
    {
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Reseting Li #%d... ", nLi);
        result = OtxDrvReset(g_hPhysDevs.hLi[nLi]); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("Done.");
      }
    }
    break;
#endif


  case 'x':
    {
      //OTX_RESULT otxRc = OTX_S_OK;
      char sqzFileName[64];

      DemoLogf("RawDataSender[%d] => RawDataReceiver[%d]. Please enter file to send: ", nCurrentLi, !nCurrentLi);
      OsaConsoleGetStr(sqzFileName);

      if ((g_hLogicDevs.mRawDataSender[nCurrentLi].pTxFile = fopen(sqzFileName, "rb")) == NULL) {
        DemoLogf("Error: Could not open file %s \n", sqzFileName);
        break;
      }

#ifdef DEMO_RAW_RECEIVER
      if ((g_hLogicDevs.mRawDataReceiver[!nCurrentLi].pRxFile = fopen("T2mRawData.Raw", "wb")) == NULL) {
        DemoLogf("Error: Could not open file %s \n", sqzFileName);
        break;
      }

      DemoLogf("Creating Raw Data Receiver[%d] Thread\n", !nCurrentLi);
#ifdef DEMO_LINK_WITH_LIBCMT
      _beginthread(DemoRawDataReceiver, 0, (void *) &g_hLogicDevs.mRawDataReceiver[!nCurrentLi].nSourceId);
#else
    	OsaThreadStartEx(DemoRawDataReceiver, (void *)&g_hLogicDevs.mRawDataReceiver[!nCurrentLi].nSourceId);
#endif
      //OsaThreadSleep(500);  // Give the receiver some time to get ready
#endif
      DemoLogf("Creating Raw Data Sender[%d] Thread\n", nCurrentLi);
#ifdef DEMO_LINK_WITH_LIBCMT
      _beginthread(DemoRawDataSender, 0, (void *) &g_hLogicDevs.mRawDataSender[nCurrentLi].nSourceId);
#else
    	OsaThreadStartEx(DemoRawDataSender, (void *)&g_hLogicDevs.mRawDataSender[nCurrentLi].nSourceId);
#endif
      
    
    }
    break;


  case 'a':
    {
#ifdef DEMO_RAW_RECEIVER
      DemoLogf("Aborting Raw Data Receiver[%d] Thread\n", !nCurrentLi);
      g_hLogicDevs.mRawDataReceiver[!nCurrentLi].bActive = OTX_FALSE;
#endif

      DemoLogf("Aborting Raw Data Sender[%d] Thread\n", nCurrentLi);
      g_hLogicDevs.mRawDataSender[nCurrentLi].bActive = OTX_FALSE;

    }
    break;

  case 'v':
    {
#ifdef DEMO_RAW_RECEIVER
      g_hLogicDevs.mRawDataReceiver[!nCurrentLi].bVerbose = !g_hLogicDevs.mRawDataReceiver[!nCurrentLi].bVerbose;
      DemoLogf("Setting verbose setting to %s for Raw Data Receiver[%d] Thread\n", (g_hLogicDevs.mRawDataReceiver[!nCurrentLi].bVerbose)?"ON":"OFF", !nCurrentLi);
#endif
    }
    break;


  case 'r':
    {

#ifdef DEMO_RAW_RECEIVER
      g_hLogicDevs.mRawDataReceiver[!nCurrentLi].pRxFile = NULL;

      DemoLogf("Creating Raw Data Receiver[%d] Thread\n", !nCurrentLi);
#ifdef DEMO_LINK_WITH_LIBCMT
      _beginthread(DemoRawDataReceiver, 0, (void *) &g_hLogicDevs.mRawDataReceiver[!nCurrentLi].nSourceId);
#else
    	OsaThreadStartEx(DemoRawDataReceiver, (void *)&g_hLogicDevs.mRawDataReceiver[!nCurrentLi].nSourceId);
#endif
#endif
    }
    break;

  case 's':
    {
      //OTX_RESULT otxRc = OTX_S_OK;
      char sqzFileName[64];

      DemoLogf("RawDataSender[%d]. Please enter file to send: ", nCurrentLi);
      OsaConsoleGetStr(sqzFileName);

      if ((g_hLogicDevs.mRawDataSender[nCurrentLi].pTxFile = fopen(sqzFileName, "rb")) == NULL) {
        DemoLogf("Error: Could not open file %s \n", sqzFileName);
        break;
      }

      DemoLogf("Creating Raw Data Sender[%d] Thread\n", nCurrentLi);
#ifdef DEMO_LINK_WITH_LIBCMT
      _beginthread(DemoRawDataSender, 0, (void *) &g_hLogicDevs.mRawDataSender[nCurrentLi].nSourceId);
#else
    	OsaThreadStartEx(DemoRawDataSender, (void *)&g_hLogicDevs.mRawDataSender[nCurrentLi].nSourceId);
#endif
    
    }
    break;



  case 'S':
    {
      DemoLogf("RawDataSender[%d]. Continously send text string \"%s\".\n", nCurrentLi, DEMO_TEST_STRING);

      g_hLogicDevs.mRawDataSender[nCurrentLi].pTxFile = NULL;

      DemoLogf("Creating Raw Data Sender[%d] Thread\n", nCurrentLi);
#ifdef DEMO_LINK_WITH_LIBCMT
      _beginthread(DemoRawDataSender, 0, (void *) &g_hLogicDevs.mRawDataSender[nCurrentLi].nSourceId);
#else
    	OsaThreadStartEx(DemoRawDataSender, (void *)&g_hLogicDevs.mRawDataSender[nCurrentLi].nSourceId);
#endif
    
    }
    break;



  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 nLi;

      for (nLi = 0; nLi < 2; nLi++) {
        DemoLogf("Line status of Li #%d is ", nLi);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[nLi], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }
    }
    break;



  case 't':
    {
      OTX_UINT8 arMsg[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
      OTX_INT32 i;
      OTX_INT32 nTaskRef = 1234; // Arbitrary value
      OTX_INT32 busyCount = 0;
      clock_t startTime, stopTime;

      if (g_cbLiHdlcSenderFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
        DemoLogf("Error: Frame Length too long\n");
        break;
      }

      for (i = 0; i < g_cbLiHdlcSenderFrame; i++) {
        arMsg[i] = (OTX_UINT8) i;
      }

      DemoLogf("Sending %d HDLC frame(s) (TS16) from Li #%d (len=%d) ...", g_coHdlcSenderFrames, nCurrentLi, g_cbLiHdlcSenderFrame);

      // Send HDLC Frame(s)
      startTime = clock();
      for (i = 0; i < g_coHdlcSenderFrames; i++) {
        result = OtxHdlcSendData(g_hLogicDevs.hHdlcSender[nCurrentLi], arMsg, g_cbLiHdlcSenderFrame, nTaskRef, OTX_TIME_INFINITY); 
        if (result == OTX_E_BUSY) {
          // Driver IO Request buffer is full. We need to retransmit this frame.
          i--;
          busyCount++;
        } else {
          DemoCheckRc(g_hLogicDevs.hHdlcSender[nCurrentLi], result);
        }
      }
      stopTime = clock();

      DemoLogf("Done (BusyCount=%d, time=%2.1f seconds).\n", busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
    }
    break;


  case 'T':
    {
      OTX_UINT8 arMsg[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
      OTX_INT32 i;
      OTX_INT32 nTaskRef = 1234; // Arbitrary value
      OTX_INT32 busyCount = 0;
      clock_t startTime, stopTime;

      if (g_cbDspHdlcSenderFrame > OD_HDLC_MAX_TX_HDLC_FRAME_LEN) {
        DemoLogf("Error: Frame Length too long\n");
        break;
      }

      for (i = 0; i < g_cbDspHdlcSenderFrame; i++) {
        arMsg[i] = (OTX_UINT8) i;
      }

      DemoLogf("Sending %d HDLC frame(s) (TS1) from DSP #%d (len=%d)...", g_coHdlcSenderFrames, nCurrentLi, g_cbDspHdlcSenderFrame);

      // Send HDLC Frame(s)
      startTime = clock();
      for (i = 0; i < g_coHdlcSenderFrames; i++) {
        result = OtxHdlcSendData(g_hLogicDevs.hDspHdlcSender[nCurrentLi], arMsg, g_cbDspHdlcSenderFrame, nTaskRef, OTX_TIME_INFINITY); 
        if (result == OTX_E_BUSY) {
          // Driver IO Request buffer is full. We need to retransmit this frame.
          i--;
          busyCount++;
        } else {
          DemoCheckRc(g_hLogicDevs.hDspHdlcSender[nCurrentLi], result);
        }
      }
      stopTime = clock();

      DemoLogf("Done (BusyCount=%d, time=%2.1f seconds).\n", busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
    }
    break;


  case '0':
    // Select Li/Dsp 0 as current
    nCurrentLi = 0;
    DemoLogf("Li/Dsp#0 Selected ....\n");
    break;

  case '1':
    // Select Li/Dsp 1 as current
    nCurrentLi = 1;
    DemoLogf("Li/Dsp#1 Selected ....\n");
    break;

  case 'q':
  case 'Q':
    return 1;
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}                      




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->law               = OTX_DATA_FORMAT_COMPAND8_U_LAW;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_PCMCIA;
  envP->nRawDataCapacity  = 64;
  envP->eProgramModule    = OTX_DSP_SPM_OTXHDLC;


  for (i = 1; i < argc; i++) {        // Scan command line arguments
    OsaStdStrUpr(argv[i]);
    if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
      if (*(argv[i]+1) == 'L') {   // Log file
        envP->logFileName = argv[i]+2;
      } else if (*(argv[i]+1) == 'B') {   // Board number
        envP->boardNo = strtoul(argv[i]+2, NULL, 0);
      } else if (*(argv[i]+1) == 'T') {   // Board Type
        if ( *(argv[i]+2) == '2') {
          envP->nBoardType = OTX_DEVICE_THOR_PCI;
        } else if ( *(argv[i]+2) == 'M') {
          envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
        
      } else if (*(argv[i]+1) == 'N') {   // Capacity of Raw Data Pipe
        envP->nRawDataCapacity = strtol(argv[i]+2, NULL, 0);
        if (envP->nRawDataCapacity >= 64) {
          if ((envP->nRawDataCapacity%64) != 0) {
            DemoLogf("Error: Raw Data Pipe capacity must be a multiple of 64, or a subchannel capacity (8, 16, or 32)\n");
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else {
          switch (envP->nRawDataCapacity) {
          case 8:
          case 16:
          case 32:
            DemoLogf("When using subchannels with this demo program, please be aware the output\n");
            DemoLogf("file produced by this program might not be identical to the input\n");
            DemoLogf("file, because the transmitted octets might not be bit-aligned with the\n");
            DemoLogf("received octets. In order to compare the output file with the input file, \n");
            DemoLogf("the output file might need to be post-processed (bit-shifted).\n\n");
            break;
          default:
            DemoLogf("Error: Valid choices for subchannel Raw Data capacity are 8, 16, 32\n");
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        }
      } else if (*(argv[i]+1) == 'C') {   // Companding Law
        if (*(argv[i]+2) == 'A') {   // A-law
          envP->law = OTX_DATA_FORMAT_COMPAND8_A_LAW;
        } else if (*(argv[i]+2) == 'U') {   // U-law
          envP->law = OTX_DATA_FORMAT_COMPAND8_U_LAW;
        } else {
          DemoLogf("Error: Invalid companding law command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else if (*(argv[i]+1) == 'P') {   // Dsp Program module
        if (*(argv[i]+2) == '1') {   // OtxSpm1.out
          envP->eProgramModule = OTX_DSP_SPM_OTXSPM1;
        } else if (*(argv[i]+2) == 'H') {   // OtxHdlc.out
          envP->eProgramModule = OTX_DSP_SPM_OTXHDLC;
        } else {
          DemoLogf("Error: Invalid DSP Program Module command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
    } else if (*(argv[i]+1) == 'M') {   // Li Mode
        if (*(argv[i]+2) == 'E') {   // E1 
          envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
        } else if (*(argv[i]+2) == 'T') {   // T1
          envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
        } else {
          DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    } else {
      DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
      DemoHelpPanel();
      OsaConsoleExit();
      exit(1);
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[2|M]      : Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI Board with Vidar-5x4-ASM\n");
  DemoLogf("                  M = Use Thor-2-PCMCIA Board\n");
  DemoLogf(" /CA          : A-law\n");
  DemoLogf(" /CU          : u-law\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /N<capacity> : Raw Data Pipe capacity in kbps (in 64kbps increments)\n");
  DemoLogf(" /P1          : Use OtxSpm1 DSP module (instead of OtxHdlc as default)\n");
  DemoLogf(" /PH          : Use OtxHdlc DSP module (default)\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;

  //EnterCriticalSection(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  //LeaveCriticalSection(&g_demoLogfCritSec);
}



int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  printf("T2MRAWDATA, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Thor-2-PCMCIA Example program for Raw Data Sender/Receiver.\n");

  //InitializeCriticalSection(&g_analyzeCritSec);
  //InitializeCriticalSection(&g_demoLogfCritSec);


  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
    exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Raw Data Pipe capacity is %d\n", g_demoEnv.nRawDataCapacity);

  if ((g_demoEnv.eProgramModule == OTX_DSP_SPM_OTXSPM1) && (g_demoEnv.nRawDataCapacity != 64)) {
    DemoLogf("Error: OTX_DSP_SPM_OTXSPM1 only supports 64kpbs pipes in this demo. Exiting.\n");
    OsaConsoleExit();
    exit(-1);
  }

  

  // Connect to the Driver
  result = OtxDrvConnectLib(); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("DemoCreateEventQueue\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCMCIA board
  DemoLogf("DemoCreatePhysicalDevices\n");
  DemoCreatePhysicalDevices();

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnablePhysicalDevices\n");
  DemoEnablePhysicalDevices();

  // Set clock source for the board 
  if (g_hPhysDevs.hVidarAsmController0) {
    // Thor-2-PCI, Thor-8-PCI
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
#if 1  // Temporary
    {
      OTX_INT32 nDsp;
      for (nDsp = 0; nDsp < DEMO_DSP_MAX; nDsp++) {
        DemoLogf("Init Bsp for Dsp#%d\n", nDsp);
        OtxDrvDbgDump(g_hPhysDevs.hDsp[nDsp], 1);
      }
    }
#endif
  } else {
    // Thor-2-PCMCIA
    result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_NA, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  }

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("DemoCreateLogicalDevices\n");
  DemoCreateLogicalDevices();

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("DemoConnectDevices\n");
  DemoConnectDevices();

  // Initialize the logical devices. A logical device must be create and connected before it can be initialized.
  DemoLogf("DemoInitializeLogicalDevices\n");
  DemoInitializeLogicalDevices();

  // Enable logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("DemoEnableLogicalDevices\n");
  DemoEnableLogicalDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(300); // Only so the following message will show on the screen

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = getch();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}

